package com.yworks.eclipse.sample;

import java.awt.Color;
import java.awt.event.MouseEvent;
import java.beans.PropertyChangeListener;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

import javax.swing.AbstractAction;
import javax.swing.ActionMap;
import javax.swing.InputMap;
import javax.swing.JComponent;
import javax.swing.SwingUtilities;

import org.eclipse.albireo.core.SwingControl;
import org.eclipse.core.resources.ResourcesPlugin;
import org.eclipse.core.runtime.IPath;
import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.jface.action.Action;
import org.eclipse.jface.dialogs.MessageDialog;
import org.eclipse.jface.preference.IPreferenceStore;
import org.eclipse.jface.util.IPropertyChangeListener;
import org.eclipse.jface.util.PropertyChangeEvent;
import org.eclipse.jface.viewers.ISelectionChangedListener;
import org.eclipse.jface.viewers.IStructuredSelection;
import org.eclipse.jface.viewers.SelectionChangedEvent;
import org.eclipse.jface.viewers.StructuredSelection;
import org.eclipse.swt.SWT;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Display;
import org.eclipse.ui.IEditorInput;
import org.eclipse.ui.IEditorSite;
import org.eclipse.ui.IPathEditorInput;
import org.eclipse.ui.IPerspectiveDescriptor;
import org.eclipse.ui.IViewPart;
import org.eclipse.ui.IWorkbenchPage;
import org.eclipse.ui.PartInitException;
import org.eclipse.ui.PlatformUI;
import org.eclipse.ui.actions.ActionFactory;
import org.eclipse.ui.dialogs.SaveAsDialog;
import org.eclipse.ui.part.EditorPart;
import org.eclipse.ui.views.contentoutline.IContentOutlinePage;
import org.eclipse.ui.views.properties.PropertySheet;

import y.base.Command;
import y.base.EdgeCursor;
import y.base.Node;
import y.base.NodeCursor;
import y.io.GMLIOHandler;
import y.io.GraphMLIOHandler;
import y.io.IOHandler;
import y.io.YGFIOHandler;
import y.util.D;
import y.view.Arrow;
import y.view.DefaultBackgroundRenderer;
import y.view.DefaultGraph2DRenderer;
import y.view.EditMode;
import y.view.Graph2D;
import y.view.Graph2DClipboard;
import y.view.Graph2DEvent;
import y.view.Graph2DListener;
import y.view.Graph2DUndoManager;
import y.view.Graph2DView;
import y.view.Graph2DViewActions;
import y.view.Graph2DViewMouseWheelZoomListener;
import y.view.HitInfo;
import y.view.PolyLineEdgeRealizer;
import y.view.Selections;
import y.view.ShapeNodeRealizer;
import y.view.ViewMode;
import y.view.hierarchy.HierarchyManager;

/**
 * The graph editor part. It can display diagrams in GML and YGF formats. The
 * editor is based on the yFiles swing component Graph2DView which is integrated
 * by the means of the SWT_AWT bridge.
 * <p>
 * The editor part also acts as a SelectionProvider for the associated outline
 * page <code>GraphOutlinePage</code> and the Properties page. The outline page
 * displays the names of all nodes and edges of the graph inside a tree view.
 * The properties page displays editable attributes of the currently selected
 * elements.
 * <p>
 * The graph editor part provides special implementations for the standard edit
 * actions Undo, Redo, Cut, Copy, Paste, Delete and Select All.
 * <p>
 * The preferences page <code>GraphEditorPreferences</code> allows to adjust
 * some settings for the editor part.
 */
public class GraphEditorPart extends EditorPart implements IPropertyChangeListener {
  private final Graph2DView view;
  private final Graph2DUndoManager undoManager;

  private String filename;
  private boolean edited;

  private IPreferenceStore prefs;

  private final SwingControlledAction undoAction = new SwingControlledAction();
  private final SwingControlledAction redoAction = new SwingControlledAction();
  private final SwingControlledAction cutAction = new SwingControlledAction();
  private final SwingControlledAction pasteAction = new SwingControlledAction();
  private final SwingControlledAction copyAction = new SwingControlledAction();
  private final SwingControlledAction deleteAction = new SwingControlledAction();

  private final GraphSelectionProvider graphSelectionProvider;
  private final GraphOutlinePage outlineView;

  private final Action selectAllAction = new Action() {
    public void run() {
      SwingUtilities.invokeLater(new Runnable() {
        public void run() {
          view.getGraph2D().firePreEvent();
          Selections.selectNodesAndSelfLoopBends(view.getGraph2D());
          view.getGraph2D().firePostEvent();
          view.getGraph2D().updateViews();
        }
      });
    }
  };

  public GraphEditorPart() {
    prefs = GraphEditorPlugin.getDefault().getPreferenceStore();
    prefs.addPropertyChangeListener(this);

    // View
    view = new Graph2DView();
    new HierarchyManager(view.getGraph2D());

    undoManager = new Graph2DUndoManager(view.getGraph2D());
    undoManager.setViewContainer(view);
    initGraphView();

    // Selection Provider
    graphSelectionProvider = new GraphSelectionProvider(view);

    // Outline View
    outlineView = new GraphOutlinePage(this);
    outlineView.addSelectionChangedListener(new ISelectionChangedListener() {
      public void selectionChanged(final SelectionChangedEvent event) {
        if (event.getSelection() instanceof IStructuredSelection) {
          final IStructuredSelection selection = (IStructuredSelection) event.getSelection();
          SwingUtilities.invokeLater(new Runnable() {
            public void run() {
              Graph2D graph = view.getGraph2D();
              graph.firePreEvent();
              graph.unselectAll();
              for (Iterator iter = selection.iterator(); iter.hasNext();) {
                Object o = iter.next();
                if (o instanceof NodeProperty) {
                  graph.setSelected(((NodeProperty) o).getNode(), true);
                } else if (o instanceof EdgeProperty) {
                  graph.setSelected(((EdgeProperty) o).getEdge(), true);
                }
              }
              view.updateView();
              graph.firePostEvent();
            }
          });
        }
      }
    });
  }

  void initGraphView() {
    ViewMode viewMode = new EditMode();

    view.addViewMode(viewMode);
    view.setOpaque(true);
    view.getCanvasComponent().addMouseWheelListener(new Graph2DViewMouseWheelZoomListener());
    view.setAntialiasedPainting(true);

    PolyLineEdgeRealizer per = new PolyLineEdgeRealizer();
    per.setTargetArrow(Arrow.STANDARD);
    view.getGraph2D().setDefaultEdgeRealizer(per);

    ShapeNodeRealizer snr = new ShapeNodeRealizer();
    snr.setShapeType(ShapeNodeRealizer.ROUND_RECT);
    snr.getLabel().setFontSize(14);
    snr.setFillColor(Color.decode("#ffcc00"));
    view.getGraph2D().setDefaultNodeRealizer(snr);

    view.getGraph2D().setBackupRealizersHandler(undoManager);
    undoManager.addUndoListener(new Graph2DUndoManager.UndoListener() {
      public void commandAdded(Command arg0) {
        setDirty();
      }
    });

    undoManager.getUndoAction().addPropertyChangeListener(new PropertyChangeListener() {
      public void propertyChange(java.beans.PropertyChangeEvent evt) {
        if (!undoManager.canUndo()) {
          setUndirty();
        }
      }
    });
    undoManager.getRedoAction().addPropertyChangeListener(new PropertyChangeListener() {
      public void propertyChange(java.beans.PropertyChangeEvent evt) {
        if (undoManager.canUndo()) {
          setDirty();
        }
      }
    });

    undoAction.setSwingAction((AbstractAction) undoManager.getUndoAction());
    redoAction.setSwingAction((AbstractAction) undoManager.getRedoAction());

    Graph2DClipboard clipboard = new Graph2DClipboard(view);
    cutAction.setSwingAction((AbstractAction) clipboard.getCutAction());
    copyAction.setSwingAction((AbstractAction) clipboard.getCopyAction());
    pasteAction.setSwingAction((AbstractAction) clipboard.getPasteAction());

    Graph2DViewActions viewActions = new Graph2DViewActions(view);
    deleteAction.setSwingAction((AbstractAction) viewActions.getDeleteSelectionAction());

    ActionMap amap = viewActions.createActionMap();
    InputMap imap = viewActions.createDefaultInputMap(amap);
    view.getCanvasComponent().setActionMap(amap);
    view.getCanvasComponent().setInputMap(JComponent.WHEN_FOCUSED, imap);

    setUndirty();
  }

  public void doSave(IProgressMonitor monitor) {
    monitor.beginTask("Saving...", 100);
    saveResourcesFile(filename);
    monitor.done();
  }

  public void doSaveAs() {
    System.out.println("SaveAs..");
    SaveAsDialog dialog = new SaveAsDialog(this.getEditorSite().getShell());
    if (dialog.open() == SaveAsDialog.OK) {
      IPath p = ResourcesPlugin.getWorkspace().getRoot().getLocation().append(dialog.getResult());
      setPartName(p.lastSegment());
      filename = p.toOSString();
      exportFile(filename);
      setUndirty();
    }
  }

  public Object getAdapter(Class adapter) {
    if (IContentOutlinePage.class.equals(adapter)) {
      return outlineView;
    }
    return super.getAdapter(adapter);
  }

  public void init(IEditorSite site, IEditorInput input) throws PartInitException {
    setSite(site);
    setInput(input);
  }

  protected void setUndirty() {
    edited = false;
    Display.getDefault().asyncExec(new Runnable() {
      public void run() {
        firePropertyChange(PROP_DIRTY);
      }
    });
  }

  public void setDirty() {
    edited = true;
    Display.getDefault().asyncExec(new Runnable() {
      public void run() {
        firePropertyChange(PROP_DIRTY);
      }
    });
  }

  public boolean isDirty() {
    return edited;
  }

  public boolean isSaveAsAllowed() {
    return true;
  }

  private SwingControl swingControl;

  public void createPartControl(final Composite parent) {
    swingControl = new SwingControl(parent, SWT.NONE) {
      protected JComponent createSwingComponent() {
        return view;
      }

      public Composite getLayoutAncestor() {
        return parent;
      }
    };

    bindActions();

    openResourcesFile(getEditorInput());

    getSite().setSelectionProvider(graphSelectionProvider);

    try {
      IPerspectiveDescriptor perspective = getSite().getWorkbenchWindow().getActivePage().getPerspective();
      if (!"com.yworks.eclipse.sample.GraphPerspective".equals(perspective.getId())) {
        if (MessageDialog.openQuestion(getSite().getShell(), "Switch perspective",
            "Switch to Graph Editing perspective?")) {
          PlatformUI.getWorkbench().showPerspective("com.yworks.eclipse.sample.GraphPerspective",
              getSite().getWorkbenchWindow());
        }
      }
    } catch (Exception ignore) {
    }

    // notify properties view on graph changes
    final Graph2D graph = getView().getGraph2D();
    addViewEditedListeners(graph);

    graph.addGraph2DListener(new Graph2DListener() {
      public void onGraph2DEvent(Graph2DEvent event) {
        if (event.getPropertyName().equals("text")) {
          notifyPropertySheet(graph);
        }
      }
    });
  }

  private void addViewEditedListeners(final Graph2D graph) {
    Iterator<ViewMode> viewModes = (Iterator<ViewMode>) getView().getViewModes();
    while (viewModes.hasNext()) {
      ViewMode viewMode = viewModes.next();
      final PropertyChangeListener listener = new PropertyChangeListener() {
        public void propertyChange(java.beans.PropertyChangeEvent evt) {
          if (evt.getNewValue() == Boolean.FALSE) {
            notifyPropertySheet(graph);
          }
        }
      };
      viewMode.addPropertyChangeListener(ViewMode.EDITING_PROPERTY, listener);

      viewMode.addPropertyChangeListener(ViewMode.CHILD_MODE_PROPERTY, new PropertyChangeListener() {
        public void propertyChange(java.beans.PropertyChangeEvent evt) {
          if (evt.getNewValue() == null) {
            ((ViewMode) evt.getOldValue()).removePropertyChangeListener(listener);
          } else {
            ((ViewMode) evt.getNewValue()).addPropertyChangeListener(listener);
          }
        }
      });
    }
  }

  private void notifyPropertySheet(final Graph2D graph) {
    final IWorkbenchPage activePage = getSite().getWorkbenchWindow().getActivePage();
    Display display = activePage.getWorkbenchWindow().getShell().getDisplay();
    if (display != null) {
      display.asyncExec(new Runnable() {
        public void run() {
          final IViewPart properties = activePage.findView("org.eclipse.ui.views.PropertySheet");
          if (properties instanceof PropertySheet) {
            PropertySheet propertySheet = (PropertySheet) properties;
            propertySheet.selectionChanged(GraphEditorPart.this, new StructuredGraphSelection(graph));
          }
        }
      });
    }
  }

  private void bindActions() {
    getEditorSite().getActionBars().setGlobalActionHandler(ActionFactory.UNDO.getId(), undoAction);
    getEditorSite().getActionBars().setGlobalActionHandler(ActionFactory.REDO.getId(), redoAction);
    getEditorSite().getActionBars().setGlobalActionHandler(ActionFactory.COPY.getId(), copyAction);
    getEditorSite().getActionBars().setGlobalActionHandler(ActionFactory.CUT.getId(), cutAction);
    getEditorSite().getActionBars().setGlobalActionHandler(ActionFactory.PASTE.getId(), pasteAction);
    getEditorSite().getActionBars().setGlobalActionHandler(ActionFactory.SELECT_ALL.getId(), selectAllAction);
    getEditorSite().getActionBars().setGlobalActionHandler(ActionFactory.DELETE.getId(), deleteAction);
  }

  private void saveResourcesFile(String filename) {
    exportFile(filename);
  }

  protected void openResourcesFile(IEditorInput input) {
    if (!(input instanceof IPathEditorInput)) {
      return;
    }

    IPathEditorInput fInput = (IPathEditorInput) input;
    final IPath path = fInput.getPath();
    setPartName(path.lastSegment());
    filename = path.toOSString();

    Display.getCurrent().asyncExec(new Runnable() {
      public void run() {
        if (importFile(path.toOSString())) {
          SwingUtilities.invokeLater(new Runnable() {
            public void run() {
              undoManager.resetQueue();
              setUndirty();
              view.fitContent();
              view.updateView();
            }
          });
        }
      }
    });
  }

  protected boolean importFile(final String file) {
    IOHandler ioh = null;
    if (file.endsWith(".gml")) {
      ioh = new GMLIOHandler();
    } else if (file.endsWith(".ygf")) {
      ioh = new YGFIOHandler();
    } else if (file.endsWith(".graphml")) {
      ioh = createGraphMLIOHandler();
    }

    if (ioh != null) {
      try {
        final IOHandler ioh1 = ioh;
        SwingUtilities.invokeAndWait(new Runnable() {
          public void run() {
            try {
              view.getGraph2D().clear();
              ioh1.read(view.getGraph2D(), file);
            } catch (IOException e) {
              D.show(e);
            }
          }
        });
        return true;
      } catch (Exception ex) {
        D.show(ex);
      }
    }
    return false;
  }

  private IOHandler createGraphMLIOHandler() {
    return new GraphMLIOHandler();
  }

  protected void exportFile(final String file) {
    IOHandler ioh = null;
    if (file.endsWith(".ygf")) {
      ioh = new YGFIOHandler();
    } else if (file.endsWith(".gml")) {
      ioh = new GMLIOHandler();
    } else if (file.endsWith(".graphml")) {
      ioh = createGraphMLIOHandler();
    }
    if (ioh != null) {
      try {
        ioh.write(view.getGraph2D(), file);
        setUndirty();
      } catch (Exception ex) {
        showMessage(ex.toString());
      }
    }
  }

  public Graph2DView getView() {
    return view;
  }

  public void setFocus() {
    swingControl.setFocus();
    bindActions();
    getSite().setSelectionProvider(graphSelectionProvider);
  }

  protected void showMessage(final String message) {
    getSite().getShell().getDisplay().syncExec(new Runnable() {
      public void run() {
        MessageDialog.openInformation(getSite().getShell(), "yFiles Graph Editor", message);
      }
    });
  }

  public void updatePreferences() {
    view.setAntialiasedPainting(prefs.getBoolean("GENERAL_ENABLE_ANTIALISING"));
    view.setPaintDetailThreshold(Double.parseDouble(prefs.getString("GENERAL_PAINT_DETAILS_THRESHOLD")));
    String c = prefs.getString("BACKGROUND_COLOR");
    if (c != null && !c.equals("")) {
      String[] col = c.split(",");
      ((DefaultBackgroundRenderer) view.getBackgroundRenderer()).setColor(new java.awt.Color(Integer.parseInt(col[0]),
          Integer.parseInt(col[1]), Integer.parseInt(col[2])));

    }
    ((DefaultGraph2DRenderer) view.getGraph2DRenderer()).setDrawEdgesFirst(prefs
        .getBoolean("GENERAL_PAINT_NODES_OVER_EDGES"));
    undoManager.setMaximumUndoDepth(prefs.getInt("GENERAL_MAXIMUM_UNDO_DEPTH"));
    view.setGridVisible(prefs.getBoolean("GRID_VISIBLE"));
    view.setGridMode(prefs.getBoolean("GRID_SNAP_TO_GRID"));
    String c2 = prefs.getString("GRID_COLOR");
    if (c2 != null && !c2.equals("")) {
      String[] col = c2.split(",");
      view
          .setGridColor(new java.awt.Color(Integer.parseInt(col[0]), Integer.parseInt(col[1]), Integer.parseInt(col[2])));
    }

    view.setGridResolution((double) prefs.getInt("GRID_SPACING"));

    updateViewLater();
  }

  private void updateViewLater() {
    SwingUtilities.invokeLater(new Runnable() {
      public void run() {
        view.updateView();
      }
    });
  }

  public void propertyChange(PropertyChangeEvent event) {
    updatePreferences();
  }

  static class StructuredGraphSelection extends StructuredSelection {
    StructuredGraphSelection(List list) {
      super(list);
    }

    public StructuredGraphSelection(Graph2D graph) {
      this(createList(graph));
    }

    private static List createList(Graph2D graph) {
      final List structureSelection = new ArrayList();
      for (NodeCursor nc = graph.nodes(); nc.ok(); nc.next()) {
        if (graph.isSelected(nc.node())) {
          structureSelection.add(new NodeProperty(nc.node()));
        }
      }
      for (EdgeCursor ec = graph.edges(); ec.ok(); ec.next()) {
        if (graph.isSelected(ec.edge())) {
          structureSelection.add(new EdgeProperty(ec.edge()));
        }
      }
      return structureSelection;
    }
  }

}
