package com.yworks.eclipse.sample;


import org.eclipse.jface.action.Action;
import org.eclipse.jface.viewers.DoubleClickEvent;
import org.eclipse.jface.viewers.IDoubleClickListener;
import org.eclipse.jface.viewers.ISelection;
import org.eclipse.jface.viewers.IStructuredSelection;
import org.eclipse.jface.viewers.ITableLabelProvider;
import org.eclipse.jface.viewers.ITreeContentProvider;
import org.eclipse.jface.viewers.LabelProvider;
import org.eclipse.jface.viewers.SelectionChangedEvent;
import org.eclipse.jface.viewers.StructuredSelection;
import org.eclipse.jface.viewers.Viewer;
import org.eclipse.swt.graphics.Image;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.ui.ISelectionListener;
import org.eclipse.ui.ISharedImages;
import org.eclipse.ui.IWorkbenchPart;
import org.eclipse.ui.PlatformUI;
import org.eclipse.ui.views.contentoutline.ContentOutlinePage;
import y.base.Edge;
import y.base.EdgeCursor;
import y.base.GraphEvent;
import y.base.GraphListener;
import y.base.Node;
import y.base.NodeCursor;
import y.view.Graph2D;
import y.view.Graph2DEvent;
import y.view.Graph2DListener;

/**
 * Outline page for a graph document. This outline page is returned as an 
 * adapter of <code>GraphEditorPart</code>. 
 */
public class GraphOutlinePage extends ContentOutlinePage {
  private Action centerNodeAction;
  private Action refreshAction;

  private ViewContentProvider viewContentProvider;

  private GraphEditorPart graphEditorPart;

  private StructuredSelection lastSelectionEvent;

  private boolean listenToGraph2DEvents = true;
  private boolean listenToGraphSelectionEvents = true;
  private boolean listenToTreeSelectionEvents = true;

  private int editBracketCount;

  private boolean editorSelectionChange;

  public GraphOutlinePage( GraphEditorPart editor ) {
    this.graphEditorPart = editor;
  }

  class ViewContentProvider implements ITreeContentProvider {
    GraphEditorPart yEditor;

    public void inputChanged( Viewer viewer, Object oldInput, Object newInput ) {
      yEditor = ( GraphEditorPart ) newInput;
    }

    public void dispose() {
      yEditor = null;
    }

    public Object[] getElements( Object inputElement ) {
      return getChildren( inputElement );
    }

    public Object[] getChildren( Object parentElement ) {      
      if ( parentElement instanceof NodeProperty ) {
        return null;
      } else if ( parentElement instanceof EdgeProperty ) {
        return null;
      } else if ( "Nodes".equals( parentElement ) ) {
        Graph2D graph = yEditor.getView().getGraph2D();
        Object[] result = new Object[graph.nodeCount()];
        int i = 0;
        for ( NodeCursor nc = graph.nodes(); nc.ok(); nc.next(), i++ ) {
          result[ i ] = new NodeProperty( nc.node() );
        }
        return result;
      } else if ( "Edges".equals( parentElement ) ) {
        Graph2D graph = yEditor.getView().getGraph2D();
        Object[] result = new Object[graph.edgeCount()];
        int i = 0;
        for ( EdgeCursor ec = graph.edges(); ec.ok(); ec.next(), i++ ) {
          result[ i ] = new EdgeProperty( ec.edge() );
        }
        return result;
      } else {
        return new Object[]{ "Nodes", "Edges" };
      }
    }

    public Object getParent( Object element ) {
      return null;
    }

    public boolean hasChildren( Object element ) {
      Graph2D graph = yEditor.getView().getGraph2D();
      if ( element instanceof NodeProperty ) {
        return false;
      } else if ( element instanceof EdgeProperty ) {
        return false;
      } else if ( "Nodes".equals( element ) ) {
        return graph.nodeCount() > 0;
      } else if ( "Edges".equals( element ) ) {
        return graph.edgeCount() > 0;
      } else {
        return true;
      }
    }
  }

  class ViewLabelProvider extends LabelProvider implements ITableLabelProvider {
    public String getColumnText( Object obj, int index ) {
      return getText( obj );
    }

    public Image getColumnImage( Object obj, int index ) {
      return getImage( obj );
    }

    public Image getImage( Object obj ) {
      return PlatformUI.getWorkbench().getSharedImages().getImage( ISharedImages.IMG_OBJ_ELEMENT );
    }
  }

  /**
   * This is a callback that will allow us
   * to create the viewer and initialize it.
   */
  public void createControl( Composite parent ) {
    super.createControl( parent );

    getTreeViewer().setContentProvider( viewContentProvider = new ViewContentProvider() );
    getTreeViewer().setLabelProvider( new ViewLabelProvider() );
    getTreeViewer().setInput( graphEditorPart );
    makeActions();
    hookDoubleClickAction();

    graphEditorPart.getSite().getPage().addSelectionListener( new ISelectionListener() {
      public void selectionChanged( IWorkbenchPart part, ISelection selection ) {
        if ( part == graphEditorPart && selection instanceof StructuredSelection ) {
          editorSelectionChange = true;
          setSelection( selection);         
          editorSelectionChange = false;
        }
      }
    } );

    graphEditorPart.getView().getGraph2D().addGraph2DListener( new Graph2DListener() {
      public void onGraph2DEvent( Graph2DEvent event ) {
        if ( getTreeViewer().getControl().isDisposed() ) return;
        if ( "text".equals( event.getPropertyName() ) ) {
          if ( editBracketCount <= 0 ) {
            update();
            editBracketCount = 0;
          }
        }
      }
    } );

    graphEditorPart.getView().getGraph2D().addGraphListener( new GraphListener() {
      public void onGraphEvent( GraphEvent event ) {
        if ( getTreeViewer().getControl().isDisposed() ) return;
        switch ( event.getType() ) {
          case GraphEvent.PRE_EVENT:
            editBracketCount++;
            break;
          case GraphEvent.POST_EVENT:
            editBracketCount--;
            break;
          default:
            graphChanged = true;
        }
        if ( editBracketCount <= 0 && graphChanged ) {
          update();
          editBracketCount = 0;
          graphChanged = false;
        }
      }
    } );
  }

  public void selectionChanged( final SelectionChangedEvent event ) {
    if ( !editorSelectionChange ) {
      super.selectionChanged( event );
    }
  }

  private void makeActions() {
    centerNodeAction = new Action() {
      public void run() {
        ISelection selection = getTreeViewer().getSelection();
        Object obj = ( ( IStructuredSelection ) selection ).getFirstElement();
        if ( obj instanceof NodeProperty ) {
          Node n = ( ( NodeProperty ) obj ).getNode();
          graphEditorPart.getView().setCenter( graphEditorPart.getView().getGraph2D().getCenterX( n ), graphEditorPart.getView().getGraph2D().getCenterY( n ) );
          graphEditorPart.getView().updateView();
        }
        if ( obj instanceof EdgeProperty ) {
          Edge e = ( ( EdgeProperty ) obj ).getEdge();
          double x = ( graphEditorPart.getView().getGraph2D().getCenterX( e.source() ) + graphEditorPart.getView().getGraph2D().getCenterX( e.target() ) ) / 2.0;
          double y = ( graphEditorPart.getView().getGraph2D().getCenterY( e.source() ) + graphEditorPart.getView().getGraph2D().getCenterY( e.target() ) ) / 2.0;
          graphEditorPart.getView().setCenter( x, y );
          graphEditorPart.getView().updateView();
        }
      }
    };
    refreshAction = new Action() {
      public void run() {
        getTreeViewer().refresh();
      }
    };
  }

  private void hookDoubleClickAction() {
    getTreeViewer().addDoubleClickListener( new IDoubleClickListener() {
      public void doubleClick( DoubleClickEvent event ) {
        centerNodeAction.run();
      }
    } );
  }

  public void update() {
    if ( getTreeViewer().getControl().isDisposed() ) return;
    getControl().getDisplay().asyncExec( new Runnable() {
      public void run() {
        refreshAction.run();
      }
    } );
  }

  boolean graphChanged;
}